package ch.unige.solidify.service;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Order;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.ResponseExtractor;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.exception.RestResponseErrorHandler;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.RestClientTool;

public abstract class SolidifyRestClientService {
  private static final Logger log = LoggerFactory.getLogger(SolidifyRestClientService.class);

  private static final int BUFFER_SIZE = 16384; // 2 ^ 14

  protected MessageService messageService;
  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final RestClientTool restClientTool;

  public SolidifyRestClientService(RestClientTool restClientTool, MessageService messageService,
          HttpRequestInfoProvider httpRequestInfoProvider) {
    this.restClientTool = restClientTool;
    this.messageService = messageService;
    this.httpRequestInfoProvider = httpRequestInfoProvider;
  }

  /**
   * @param <T>
   * @param url
   * @param clazz
   * @param pageable
   * @return
   */

  public <T> Collection<T> getResourceList(String url, Class<T> clazz, Pageable pageable) {
    final String jsonResult = this.getResources(url, pageable);
    final List<T> list = CollectionTool.getList(jsonResult, clazz);
    final CollectionPage page = CollectionTool.getPage(jsonResult);
    return new Collection<>(list, page);
  }

  /**
   * Complete an URL with pagination query parameters
   *
   * @param url      url to complete
   * @param pageable pageable object used to complete the URL
   * @return completed URL
   */
  protected String completeUrlWithPagination(String url, Pageable pageable) {
    final UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);

    if (pageable != null) {
      final MultiValueMap<String, String> existingQueryParams = new LinkedMultiValueMap<>(uriBuilder.build().getQueryParams());
      existingQueryParams.remove("page");
      existingQueryParams.remove("size");
      existingQueryParams.remove("sort");

      final int page = pageable.getPageNumber();
      final int size = pageable.getPageSize();
      final Sort sort = pageable.getSort();

      final MultiValueMap<String, String> pageParams = new LinkedMultiValueMap<>();
      pageParams.set("page", Integer.toString(page));
      pageParams.set("size", Integer.toString(size));

      // TODO: support sorting on many fields ?
      for (final Order order : sort) {
        final String sortProperty = order.getProperty();
        final String sortDirection = order.getDirection().toString();
        pageParams.set("sort", sortProperty + "," + sortDirection);
      }
      uriBuilder.replaceQueryParams(existingQueryParams).queryParams(pageParams);
    }
    return uriBuilder.build().toString();
  }

  // **********************
  // ** DELETE Functions **
  // **********************

  protected void deleteResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.DELETE);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {
      restClt.exchange(url, HttpMethod.DELETE, httpEntity, Void.class);
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected void deleteResourceList(String url, List<String> ids) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.DELETE);
    final HttpEntity<?> httpEntity = new HttpEntity<>(ids, this.getIncomingRequestAuthorizationHeader());
    try {
      restClt.exchange(url, HttpMethod.DELETE, httpEntity, String.class);
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  protected MultiValueMap<String, String> getIncomingRequestAuthorizationHeader() {
    return this.httpRequestInfoProvider.getAuthorizationHeader();
  }

  // *******************
  // ** GET Functions **
  // *******************

  protected String getResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {
      final ResponseEntity<String> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  public <T> List<T> getResourceList(String url, Class<T[]> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {
      final ResponseEntity<T[]> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, clazz);
      return Arrays.asList(responseEntity.getBody());

    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected <T> T getResource(String url, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected String getResource(String url, Pageable pageable) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());

    UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);
    // Add Pagination info
    if (pageable != null) {
      uriBuilder.queryParam(ActionName.PAGE, pageable.getPageNumber());
      uriBuilder.queryParam(ActionName.SIZE, pageable.getPageSize());
      // Add Sort info
      pageable.getSort()
              .stream()
              .forEach(s -> uriBuilder.queryParam(ActionName.SORT, s.getProperty() + "," + s.getDirection().toString().toLowerCase()));
    }
    URI uri = uriBuilder.build(true).encode().toUri();

    try {
      this.logModuleCall(url);
      final ResponseEntity<String> responseEntity = restClt.exchange(uri, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected String getResources(String url, Pageable pageable) {
    url = this.completeUrlWithPagination(url, pageable);

    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {

      this.logModuleCall(url);

      return restClt.exchange(url, HttpMethod.GET, httpEntity, String.class).getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected RestTemplate getRestTemplate(String url, HttpMethod httpMethod) {
    final RestTemplate restClt = this.restClientTool.getClient(url);
    switch (httpMethod) {
      case GET:
      case HEAD:
        break;
      default:
        restClt.setErrorHandler(new RestResponseErrorHandler());
    }
    return restClt;
  }

  // ********************
  // ** HEAD Functions **
  // ********************

  protected HttpStatus headResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.HEAD);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {
      restClt.exchange(url, HttpMethod.HEAD, httpEntity, Void.class);
    } catch (final HttpClientErrorException e) {
      return e.getStatusCode();
    }
    return HttpStatus.OK;
  }

  /**
   * Perform a HEAD request and return the Content-Length header value
   *
   * @param url The url of the resource to get Content-Length from
   * @return The value of the Content-Length header
   */
  public long headContentLength(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.HEAD);
    MultiValueMap<String, String> requestHeaders = this.getIncomingRequestAuthorizationHeader();
    // explicitly tell server to disable encoding to have more probability to get a Content-Length value
    requestHeaders.add("Accept-Encoding", "");
    final HttpEntity<?> httpEntity = new HttpEntity<>(requestHeaders);
    try {
      HttpEntity<Void> response = restClt.exchange(url, HttpMethod.HEAD, httpEntity, Void.class);
      HttpHeaders headers = response.getHeaders();
      List<String> lengths = headers.get("Content-Length");
      if (lengths != null && lengths.size() == 1) {
        return Long.parseLong(lengths.get(0));
      } else if (lengths != null && lengths.size() > 1) {
        throw new SolidifyRuntimeException("too many 'Content-Length' headers received");
      } else {
        throw new SolidifyRuntimeException("No 'Content-Length' headers received");
      }
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  // ********************
  // ** PATCH Function **
  // ********************

  protected <T> T patchResource(String url, T updatedResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.PATCH);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    final HttpEntity<?> httpEntity = new HttpEntity<>(updatedResource, this.getIncomingRequestAuthorizationHeader());
    try {
      return restClt.exchange(url, HttpMethod.PATCH, httpEntity, clazz).getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  // ********************
  // ** POST Functions **
  // ********************

  protected <T> T postResource(String url, Class<T> clazz) {
    final RestTemplate restClt = this.restClientTool.getClient(url);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getIncomingRequestAuthorizationHeader());
    try {

      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();

    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected <T> T postResource(String url, T newResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.POST);
    final HttpEntity<?> httpEntity = new HttpEntity<>(newResource, this.getIncomingRequestAuthorizationHeader());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  protected Result postResourceAction(String url) {
    return this.postResource(url, Result.class);
  }

  protected <T, V> T postResourceOther(String url, V otherResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.POST);
    final HttpEntity<?> httpEntity = new HttpEntity<>(otherResource, this.getIncomingRequestAuthorizationHeader());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      throw this.getSolidifyExceptionFromHttpClientErrorException(url, e);
    }
  }

  // *********************
  // ** Misc. Functions **
  // *********************

  protected <T> T uploadContent(String url, org.springframework.core.io.Resource resource, String parameterName, Class<T> clazz) {

    final MultiValueMap<String, Object> map = new LinkedMultiValueMap<>();

    map.add(parameterName, resource);

    final HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);

    final HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.PATCH);
    final ResponseEntity<T> res = restClt.exchange(url, HttpMethod.POST, requestEntity, clazz);

    return res.getBody();
  }

  protected ByteArrayOutputStream downloadContent(String url) {
    final ResponseExtractor<ByteArrayOutputStream> responseExtractor = response -> {
      int nRead;
      final byte[] data = new byte[BUFFER_SIZE];
      try (final InputStream inputStream = response.getBody();
              final ByteArrayOutputStream buffer = new ByteArrayOutputStream()) {
        while ((nRead = inputStream.read(data, 0, data.length)) != -1) {
          buffer.write(data, 0, nRead);
        }
        buffer.flush();
        return buffer;
      } catch (final IOException e) {
        throw new SolidifyRuntimeException("Impossible to download content at " + url, e);
      }
    };
    return this.downloadContent(url, responseExtractor);
  }

  public <T> T downloadContent(String url, ResponseExtractor<T> responseExtractor) {
    return this.restClientTool.downloadContent(URI.create(url), responseExtractor);
  }

  private String getResourceErrorMessage(String url) {
    return "Error in getting resource (" + url + ")";
  }

  private void logModuleCall(String moduleUrl) {
    try {
      final URL url = new URL(moduleUrl);
      final String userInfo = url.getUserInfo();
      if (userInfo != null) {
        final String userInfoWithoutPassword = userInfo.substring(0, userInfo.indexOf(':')) + ":**********";
        final String urlWithoutPassword = url.getProtocol() + "://" + userInfoWithoutPassword + "@" + url.getHost() + url
                .getPort() + url.getPath() + "?" + url.getQuery();
        log.info("Module call GET: {}", urlWithoutPassword);
      } else {
        log.info("Module call GET: {}", moduleUrl);
      }
    } catch (final MalformedURLException e) {
      log.error("Malformed url : {}", moduleUrl, e);
    }
  }

  private SolidifyRuntimeException getSolidifyExceptionFromHttpClientErrorException(String url, HttpClientErrorException e) {
    if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
      return new SolidifyResourceNotFoundException(url);
    }
    final String message = this.getResourceErrorMessage(url);
    return new SolidifyRestException(message, e);
  }
}
